(function (global, factory) {
  if (typeof define === "function" && define.amd) {
    define(['exports', './Nymph', './Entity'], factory);
  } else if (typeof exports !== "undefined") {
    factory(exports, require('./Nymph'), require('./Entity'));
  } else {
    var mod = {
      exports: {}
    };
    factory(mod.exports, global.Nymph, global.Entity);
    global.PubSub = mod.exports;
  }
})(this, function (exports, _Nymph, _Entity) {
  /* global WebSocket */
  'use strict';

  Object.defineProperty(exports, "__esModule", {
    value: true
  });
  exports.PubSubSubscription = exports.PubSub = undefined;

  function _classCallCheck(instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  }

  var _createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  if (typeof WebSocket === 'undefined') {
    throw new Error('Nymph-PubSub requires WebSocket!');
  }

  var PubSub = exports.PubSub = function () {
    function PubSub() {
      _classCallCheck(this, PubSub);
    }

    _createClass(PubSub, null, [{
      key: 'init',

      // === Static Methods ===

      value: function init(NymphOptions) {
        this.pubsubURL = NymphOptions.pubsubURL;
        if (NymphOptions.rateLimit) {
          this.rateLimit = NymphOptions.rateLimit;
        }

        this.connect();

        // Override the original Nymph methods to allow subscriptions.
        var getEntities = _Nymph.Nymph.getEntities;
        var getEntity = _Nymph.Nymph.getEntity;
        var getUID = _Nymph.Nymph.getUID;

        _Nymph.Nymph.getEntities = function (options) {
          for (var _len = arguments.length, selectors = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
            selectors[_key - 1] = arguments[_key];
          }

          var promise = getEntities.apply(_Nymph.Nymph, [options].concat(selectors));
          promise.query = JSON.stringify([options].concat(selectors));
          promise.subscribe = function (resolve, reject, count) {
            var callbacks = [resolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeQuery(promise.query, callbacks);
            return new PubSubSubscription(promise.query, callbacks, function () {
              PubSub.unsubscribeQuery(promise.query, callbacks);
            });
          };
          return promise;
        };

        _Nymph.Nymph.getEntity = function (options) {
          for (var _len2 = arguments.length, selectors = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
            selectors[_key2 - 1] = arguments[_key2];
          }

          var promise = getEntity.apply(_Nymph.Nymph, [options].concat(selectors));
          options.limit = 1;
          promise.query = JSON.stringify([options].concat(selectors));
          promise.subscribe = function (resolve, reject, count) {
            var newResolve = function newResolve(args) {
              if (!args.length) {
                if (resolve) {
                  resolve(null);
                }
              } else {
                if (resolve) {
                  resolve(args[0]);
                }
              }
            };
            var callbacks = [newResolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeQuery(promise.query, callbacks);
            return new PubSubSubscription(promise.query, callbacks, function () {
              PubSub.unsubscribeQuery(promise.query, callbacks);
            });
          };
          return promise;
        };

        _Nymph.Nymph.getUID = function (name) {
          var promise = getUID.apply(_Nymph.Nymph, [name]);
          promise.subscribe = function (resolve, reject, count) {
            var callbacks = [resolve, reject, count];

            promise.then(resolve, reject);

            PubSub.subscribeUID(name, callbacks);
            return {
              unsubscribe: function unsubscribe() {
                PubSub.unsubscribeUID(name, callbacks);
              }
            };
          };
          return promise;
        };

        _Entity.Entity.prototype.subscribe = function (resolve, reject, count) {
          var _this = this;

          if (!this.guid) {
            return false;
          }
          var query = [{ 'class': this.constructor.class, 'limit': 1 }, { type: '&', guid: this.guid }];
          var jsonQuery = JSON.stringify(query);

          var newResolve = function newResolve(args) {
            if (!args.length) {
              _this.guid = null;
              if (resolve) {
                resolve(_this);
              }
            } else {
              _this.init(args[0]);
              if (resolve) {
                resolve(_this);
              }
            }
          };
          var callbacks = [newResolve, reject, count];

          PubSub.subscribeQuery(jsonQuery, callbacks);
          return {
            unsubscribe: function unsubscribe() {
              PubSub.unsubscribeQuery(jsonQuery, callbacks);
            }
          };
        };

        return this;
      }
    }, {
      key: 'connect',
      value: function connect() {
        var _this2 = this;

        var timedAttemptConnect = function timedAttemptConnect() {
          // Attempt to connect, wait 5 seconds, then check and attempt again if unsuccessful.
          setTimeout(function () {
            if (_this2.connection.readyState !== WebSocket.OPEN) {
              _this2.connection.close();
              timedAttemptConnect();
            }
          }, 5000);

          _this2.connection = new WebSocket(_this2.pubsubURL);

          _this2.connection.onopen = function () {
            if (typeof console !== 'undefined') {
              console.log('Nymph-PubSub connection established!');
            }

            for (var query in _this2.subscriptions.queries) {
              if (!_this2.subscriptions.queries.hasOwnProperty(query)) {
                continue;
              }
              var count = false;
              for (var callbacks = 0; callbacks < _this2.subscriptions.queries[query].length; callbacks++) {
                if (typeof _this2.subscriptions.queries[query][callbacks][2] !== 'undefined') {
                  count = true;
                  break;
                }
              }
              _this2._subscribeQuery(query, count);
            }

            for (var name in _this2.subscriptions.uids) {
              if (!_this2.subscriptions.uids.hasOwnProperty(name)) {
                continue;
              }
              var _count = false;
              for (var _callbacks = 0; _callbacks < _this2.subscriptions.uids[name].length; _callbacks++) {
                if (typeof _this2.subscriptions.uids[name][_callbacks][2] !== 'undefined') {
                  _count = true;
                  break;
                }
              }
              _this2._subscribeUID(name, _count);
            }

            _this2.connection.onclose = function (e) {
              if (typeof console !== 'undefined') {
                console.log('Nymph-PubSub connection closed: ', e);
              }
              if (e.code !== 1000) {
                _this2.connection.close();
                timedAttemptConnect();
              }
            };
          };

          _this2.connection.onmessage = function (e) {
            var func = void 0,
                data = void 0;
            if (!_this2.rateLimit || typeof _this2.debouncers[e.data] === 'undefined') {
              func = function func() {
                data = JSON.parse(e.data);
                if (typeof data.query !== 'undefined' && typeof _this2.subscriptions.queries[data.query] !== 'undefined') {
                  if (typeof data.count !== 'undefined') {
                    for (var i1 = 0; typeof _this2.subscriptions.queries[data.query] !== 'undefined' && i1 < _this2.subscriptions.queries[data.query].length; i1++) {
                      if (typeof _this2.subscriptions.queries[data.query][i1][2] !== 'undefined') {
                        _this2.subscriptions.queries[data.query][i1][2](data.count);
                      }
                    }
                  } else {
                    _Nymph.Nymph.getEntities.apply(_Nymph.Nymph, JSON.parse(data.query)).then(function () {
                      for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
                        args[_key3] = arguments[_key3];
                      }

                      for (var i = 0; typeof _this2.subscriptions.queries[data.query] !== 'undefined' && i < _this2.subscriptions.queries[data.query].length; i++) {
                        _this2.subscriptions.queries[data.query][i][0].apply(null, args);
                      }
                    }, function () {
                      for (var _len4 = arguments.length, args = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
                        args[_key4] = arguments[_key4];
                      }

                      for (var i = 0; typeof _this2.subscriptions.queries[data.query] !== 'undefined' && i < _this2.subscriptions.queries[data.query].length; i++) {
                        _this2.subscriptions.queries[data.query][i][1].apply(null, args);
                      }
                    });
                  }
                }
                if (typeof data.uid !== 'undefined' && typeof _this2.subscriptions.uids[data.uid] !== 'undefined') {
                  if (typeof data.count !== 'undefined') {
                    for (var i2 = 0; typeof _this2.subscriptions.uids[data.uid] !== 'undefined' && i2 < _this2.subscriptions.uids[data.uid].length; i2++) {
                      if (typeof _this2.subscriptions.uids[data.uid][i2][2] !== 'undefined') {
                        _this2.subscriptions.uids[data.uid][i2][2](data.count);
                      }
                    }
                  } else {
                    _Nymph.Nymph.getUID(data.uid).then(function () {
                      for (var _len5 = arguments.length, args = Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
                        args[_key5] = arguments[_key5];
                      }

                      for (var i = 0; typeof _this2.subscriptions.uids[data.uid] !== 'undefined' && i < _this2.subscriptions.uids[data.uid].length; i++) {
                        _this2.subscriptions.uids[data.uid][i][0].apply(null, args);
                      }
                    }, function () {
                      for (var _len6 = arguments.length, args = Array(_len6), _key6 = 0; _key6 < _len6; _key6++) {
                        args[_key6] = arguments[_key6];
                      }

                      for (var i = 0; typeof _this2.subscriptions.uids[data.uid] !== 'undefined' && i < _this2.subscriptions.uids[data.uid].length; i++) {
                        _this2.subscriptions.uids[data.uid][i][1].apply(null, args);
                      }
                    });
                  }
                }
                if (_this2.rateLimit) {
                  delete _this2.debouncers[e.data];
                }
              };
            }
            /* jshint -W038 */
            if (!_this2.rateLimit) {
              func();
              return;
            }
            if (typeof _this2.debouncers[e.data] === 'undefined') {
              _this2.debouncers[e.data] = _this2.debounce(func);
            }
            /* jshint +W038 */
            _this2.debouncers[e.data]();
          };
        };
        timedAttemptConnect();
      }
    }, {
      key: 'debounce',
      value: function debounce(func, immediate) {
        var timeout = void 0;
        var that = this;
        return function () {
          for (var _len7 = arguments.length, args = Array(_len7), _key7 = 0; _key7 < _len7; _key7++) {
            args[_key7] = arguments[_key7];
          }

          var context = this;
          var later = function later() {
            timeout = null;
            if (!immediate) {
              func.apply(context, args);
            }
          };
          var callNow = immediate && !timeout;
          clearTimeout(timeout);
          timeout = setTimeout(later, that.rateLimit);
          if (callNow) {
            func.apply(context, args);
          }
        };
      }
    }, {
      key: 'subscribeQuery',
      value: function subscribeQuery(query, callbacks) {
        var isNewSubscription = false;
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          this.subscriptions.queries[query] = [];
          isNewSubscription = true;
        }
        var isCountSubscribed = isNewSubscription || this._isCountSubscribedQuery(query);
        this.subscriptions.queries[query].push(callbacks);
        if (this.connection.readyState === WebSocket.OPEN) {
          if (isNewSubscription) {
            this._subscribeQuery(query, typeof callbacks[2] !== 'undefined');
          } else if (!isCountSubscribed && callbacks[2] !== 'undefined') {
            this._unsubscribeQuery(query);
            this._subscribeQuery(query, true);
          }
        }
      }
    }, {
      key: 'subscribeUID',
      value: function subscribeUID(name, callbacks) {
        var isNewSubscription = false;
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          this.subscriptions.uids[name] = [];
          isNewSubscription = true;
        }
        var isCountSubscribed = isNewSubscription || this._isCountSubscribedUID(name);
        this.subscriptions.uids[name].push(callbacks);
        if (this.connection.readyState === WebSocket.OPEN) {
          if (isNewSubscription) {
            this._subscribeUID(name, typeof callbacks[2] !== 'undefined');
          } else if (!isCountSubscribed && callbacks[2] !== 'undefined') {
            this._unsubscribeUID(name);
            this._subscribeUID(name, true);
          }
        }
      }
    }, {
      key: '_subscribeQuery',
      value: function _subscribeQuery(query, count) {
        this.connection.send(JSON.stringify({
          'action': 'subscribe',
          'query': query,
          'count': count
        }));
      }
    }, {
      key: '_subscribeUID',
      value: function _subscribeUID(name, count) {
        this.connection.send(JSON.stringify({
          'action': 'subscribe',
          'uid': name,
          'count': count
        }));
      }
    }, {
      key: '_isCountSubscribedQuery',
      value: function _isCountSubscribedQuery(query) {
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          return false;
        }
        for (var callbacks = 0; callbacks < this.subscriptions.queries[query].length; callbacks++) {
          if (typeof this.subscriptions.queries[query][callbacks][2] !== 'undefined') {
            return true;
          }
        }
        return false;
      }
    }, {
      key: '_isCountSubscribedUID',
      value: function _isCountSubscribedUID(name) {
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          return false;
        }
        for (var callbacks = 0; callbacks < this.subscriptions.uids[name].length; callbacks++) {
          if (typeof this.subscriptions.uids[name][callbacks][2] !== 'undefined') {
            return true;
          }
        }
        return false;
      }
    }, {
      key: 'unsubscribeQuery',
      value: function unsubscribeQuery(query, callbacks) {
        if (typeof this.subscriptions.queries[query] === 'undefined') {
          return;
        }
        var idx = this.subscriptions.queries[query].indexOf(callbacks);
        if (idx === -1) {
          return;
        }
        this.subscriptions.queries[query].splice(idx, 1);
        if (!this.subscriptions.queries[query].length) {
          delete this.subscriptions.queries[query];
          if (this.connection.readyState === WebSocket.OPEN) {
            this._unsubscribeQuery(query);
          }
        }
      }
    }, {
      key: 'unsubscribeUID',
      value: function unsubscribeUID(name, callbacks) {
        if (typeof this.subscriptions.uids[name] === 'undefined') {
          return;
        }
        var idx = this.subscriptions.uids[name].indexOf(callbacks);
        if (idx === -1) {
          return;
        }
        this.subscriptions.uids[name].splice(idx, 1);
        if (!this.subscriptions.uids[name].length) {
          delete this.subscriptions.uids[name];
          if (this.connection.readyState === WebSocket.OPEN) {
            this._unsubscribeUID(name);
          }
        }
      }
    }, {
      key: '_unsubscribeQuery',
      value: function _unsubscribeQuery(query) {
        this.connection.send(JSON.stringify({
          'action': 'unsubscribe',
          'query': query
        }));
      }
    }, {
      key: '_unsubscribeUID',
      value: function _unsubscribeUID(name) {
        this.connection.send(JSON.stringify({
          'action': 'unsubscribe',
          'uid': name
        }));
      }
    }]);

    return PubSub;
  }();

  // === Static Properties ===
  PubSub.connection = null;
  PubSub.pubsubURL = null;
  PubSub.rateLimit = null;
  PubSub.debouncers = {};
  PubSub.subscriptions = {
    queries: {},
    uids: {}
  };

  var PubSubSubscription = exports.PubSubSubscription = function PubSubSubscription(query, callbacks, unsubscribe) {
    _classCallCheck(this, PubSubSubscription);

    // === Instance Properties ===

    this.query = query;
    this.callbacks = callbacks;
    this.unsubscribe = unsubscribe;
  };

  if (typeof window !== 'undefined' && typeof window.NymphOptions !== 'undefined' && typeof window.NymphOptions.pubsubURL !== 'undefined') {
    PubSub.init(window.NymphOptions);
  }

  exports.default = PubSub;
});